# -*- coding:utf-8 -*-

import logging
import unittest

from portfolio import Portfolio

logging.root.setLevel(logging.ERROR)


class TestPortfolio(unittest.TestCase):

    def test_update_value_percent(self):
        p = Portfolio()
        p.update_value_percent(0)
        self.assertEqual(0, p.value_percent)
        p.market_value = 10
        p.update_value_percent(100.0)
        self.assertEqual(0.1, p.value_percent)

    def test_fee(self):
        p = Portfolio()
        self.assertEqual(5.0, p._buy_fee(0))
        self.assertEqual(5.0, p._buy_fee(1000))
        self.assertEqual(10.0, p._buy_fee(10000))
        self.assertEqual(0, p._sell_fee(0))
        self.assertEqual(1.5, p._sell_fee(1000))

    def test_buy(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015,
            min_commission=5.0, round_lot=100)
        self.assertEqual((-10010.0, 10.0, 1000),
                         p.buy(price=10.0, volume=1000))
        self.assertEqual(10.0, p.transaction_cost)
        self.assertEqual(10.01, p.avg_price)
        self.assertEqual(10.0, p._price)
        self.assertEqual(1000, p.volume)
        self.assertEqual(1000, p.frozen_volume)
        self.assertEqual(10.0, p.all_transaction_cost)
        self.assertEqual((-10010.0, 10.0, 1000),
                         p.buy(price=10.0, volume=1000))
        self.assertEqual(20.0, p.transaction_cost)
        self.assertEqual(10.01, p.avg_price)
        self.assertEqual(10.0, p._price)
        self.assertEqual(2000, p.volume)
        self.assertEqual(2000, p.frozen_volume)
        self.assertEqual(20.0, p.all_transaction_cost)

    def test_update_befor_trade(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015,
            min_commission=5.0, round_lot=100)
        p.buy(price=10.0, volume=1000)
        p.update_before_trade(divide_rate=1.0)
        self.assertEqual(1000, p.sellable)
        self.assertEqual(0, p.frozen_volume)
        self.assertEqual(0, p.transaction_cost)

        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0, round_lot=100,
            divide_rate_threshold=1.005)
        p.buy(price=10.0, volume=1000)
        p.update_before_trade(divide_rate=1.1)
        self.assertEqual(1100, p.volume)
        self.assertEqual(1100, p.sellable)
        # 有拆分
        p.update_before_trade(divide_rate=1.006)
        self.assertEqual(1106, p.volume)
        self.assertEqual(1106, p.sellable)
        # 没有拆分
        p.update_before_trade(divide_rate=1.005)
        self.assertEqual(1106, p.volume)
        self.assertEqual(1106, p.sellable)

    def test_sell(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0, round_lot=100)
        self.assertEqual((-10010.0, 10.0, 1000),
                         p.buy(price=10.0, volume=1000))
        p.update_before_trade(divide_rate=1.0)
        p.sell(price=11, volume=500)
        self.assertEqual(8.25, p.transaction_cost)
        self.assertEqual(9.0365, p.avg_price)
        self.assertEqual(11.0, p._price)
        self.assertEqual(500, p.volume)
        self.assertEqual(500, p.sellable)
        self.assertEqual(18.25, p.all_transaction_cost)
        p.sell(price=11, volume=500)
        self.assertEqual(16.5, p.transaction_cost)
        self.assertEqual(0, p.avg_price)
        self.assertEqual(11, p._price)
        self.assertEqual(0, p.volume)
        self.assertEqual(0, p.sellable)
        self.assertEqual(26.5, p.all_transaction_cost)

    def test_submit_order(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0, round_lot=100)
        self.assertEqual((-10010.0, 10.0, 1000),
                         p._submit_order(side="buy", price=10.0, volume=1000))
        p.update_before_trade(divide_rate=1.0)
        self.assertEqual((5491.75, 11, 500),
                         p._submit_order(side="sell", price=11, volume=500))
        self.assertEqual(8.25, p.transaction_cost)
        self.assertEqual(9.0365, p.avg_price)
        self.assertEqual(500, p.volume)
        self.assertEqual(18.25, p.all_transaction_cost)

    def test_sell_all_stock(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0, round_lot=100)
        # sellable==0 , 无可卖
        self.assertEqual((0.0, 10, 0), p._sell_all_stock(price=10))
        self.assertEqual((-10010.0, 10.0, 1000),
                         p._submit_order(side="buy", price=10.0, volume=1000))
        p.update_before_trade(divide_rate=1.0)
        self.assertEqual((10983.5, 11, 1000), p._sell_all_stock(price=11))
        self.assertEqual(16.5, p.transaction_cost)
        self.assertEqual(0, p.avg_price)
        self.assertEqual(0, p.volume)
        self.assertEqual(26.5, p.all_transaction_cost)

    def test_order_value(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0, round_lot=100)
        # buy
        self.assertEqual((-10010.0, 10, 1000),
                         p.order_value(amount=10000, price=10,
                                       current_cash=30000))
        self.assertEqual(10.0, p.transaction_cost)
        self.assertEqual(10.01, p.avg_price)
        self.assertEqual(1000, p.volume)
        self.assertEqual(1000, p.frozen_volume)
        self.assertEqual(10.0, p.all_transaction_cost)
        p.update_before_trade(divide_rate=1.0)
        # buy 超过 cash
        self.assertEqual((-19019.0, 10, 1900),
                         p.order_value(amount=20000, price=10,
                                       current_cash=19990))
        self.assertEqual(19.0, p.transaction_cost)
        self.assertEqual(10.01, p.avg_price)
        self.assertEqual(2900, p.volume)
        self.assertEqual(1900, p.frozen_volume)
        self.assertEqual(29.0, p.all_transaction_cost)
        # sell
        self.assertEqual((4992.5, 10, 500),
                         p.order_value(amount=-5000,
                         price=10, current_cash=971.0))
        self.assertEqual(26.5, p.transaction_cost)
        self.assertEqual(10.015, round(p.avg_price, 3))
        self.assertEqual(2400, p.volume)
        self.assertEqual(1900, p.frozen_volume)
        self.assertEqual(36.5, p.all_transaction_cost)
        # sell超过 sellable
        self.assertEqual((4992.5, 10, 500),
                         p.order_value(amount=-50000, price=10,
                                       current_cash=0.0))
        self.assertEqual(34.0, p.transaction_cost)
        self.assertEqual(10.023, round(p.avg_price, 3))
        self.assertEqual(1900, p.volume)
        self.assertEqual(1900, p.frozen_volume)
        self.assertEqual(44.0, p.all_transaction_cost)

    def test_order_basic(self):
        p = Portfolio(
            buy_commission_rate=0.001,
            sell_commission_rate=0.0015, min_commission=5.0,
            round_lot=100)
        p.update_before_trade(divide_rate=1.0)
        self.assertEqual((-10010.0, 10.0, 1000),
                         p._submit_order(side="buy", price=10.0, volume=1000))
        p.update_after_trade(close_price=10.0, cash_change=-10010,
                             pre_portfolio_value=30000.0)
        # test update_after_trade
        self.assertEqual(-10.0, p.daily_pnl)
        self.assertEqual(10000.0, p.market_value)
        self.assertEqual(-10.0, p.pnl)
        self.assertEqual(0.0, round(p.daily_return, 3))
        p.update_before_trade(divide_rate=1.0)

        # sell && buy
        # pecent = 0
        self.assertEqual((10983.5, 11, 1000),
                         p.order_target_percent(
                             percent=0, price=11,
                             pre_portfolio_value=29990.0,
                             current_cash=19990))
        self.assertEqual(16.5, p.transaction_cost)
        self.assertEqual(0, p.avg_price)
        self.assertEqual(0, p.volume)
        self.assertEqual(26.5, p.all_transaction_cost)
        # percent 0.5, 增仓
        self.assertEqual((-14014.0, 10, 1400),
                         p.order_target_percent(percent=0.5, price=10,
                                                pre_portfolio_value=29990.0,
                                                current_cash=30973.5))
        self.assertEqual(30.5, p.transaction_cost)
        self.assertEqual(10.01, p.avg_price)
        self.assertEqual(1400, p.volume)
        self.assertEqual(40.5, p.all_transaction_cost)
        # test update after trade
        p.update_after_trade(close_price=10.0, cash_change=10983.5-14014.0,
                             pre_portfolio_value=29990.0)
        self.assertEqual(969.5, p.daily_pnl)
        self.assertEqual(14000.0, p.market_value)
        self.assertEqual(959.5, p.pnl)
        self.assertEqual(0.03233, round(p.daily_return, 5))
        # next day, buy_hold, then sell
        p.update_before_trade(divide_rate=1.0)
        # percent 0.452, 保持不变
        p.order_target_percent(percent=0.452, price=10,
                               pre_portfolio_value=29990.0)
        self.assertEqual(0, p.daily_pnl)
        self.assertEqual(14000.0, p.market_value)
        self.assertEqual(959.5, p.pnl)
        self.assertEqual(0.0, round(p.daily_return, 5))
        p.update_after_trade(close_price=10.0, cash_change=0.0,
                             pre_portfolio_value=29990.0)
        # next day
        p.update_before_trade(divide_rate=1.0)
        # percent 0.35, 减仓700
        p.order_target_percent(percent=0.35, price=10,
                               pre_portfolio_value=20000.0)
        self.assertEqual(10.5, p.transaction_cost)
        self.assertEqual(10.035, p.avg_price)
        self.assertEqual(700, p.volume)
        self.assertEqual(51, p.all_transaction_cost)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

import datetime
import logging
import os
import random
import unittest

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
from pandas.plotting import register_matplotlib_converters

from tenvs.envs.average import AverageEnv
from tenvs.market import Market

logging.root.setLevel(logging.INFO)
register_matplotlib_converters()


class TestAverage(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # NOTE: 需要在环境变量中设置 TUSHARE_TOKEN
        ts_token = os.getenvs("TUSHARE_TOKEN")
        self.start = "20190101"
        self.end = "20200101"
        self.codes = ["000001.SZ", "000002.SZ"]
        # self.indexs = ["000001.SH", "399001.SZ"]
        self.indexs = []
        self.show_plot = False
        # self.indexs = []
        self.data_dir = "/tmp/tenvs"
        self.m = Market(
            ts_token=ts_token,
            start=self.start,
            end=self.end,
            codes=self.codes,
            indexs=self.indexs,
            data_dir=self.data_dir)
        self.invesment = 100000.0
        self.look_back_days = 10
        self.env = AverageEnv(
            self.m,
            investment=self.invesment,
            look_back_days=self.look_back_days)

    def plot_portfolio_value(self, name):
        plt.figure(figsize=(15, 7))
        MTFmt = '%Y%m%d'
        ax = plt.gca()
        ax.xaxis.set_major_formatter(mdates.DateFormatter(MTFmt))
        plt.title(name, fontsize=10)
        dates = [datetime.datetime.strptime(d, MTFmt) for d in self.env.dates[
            self.look_back_days:]]
        plt.plot(dates,
                 self.env.portfolio_value_logs,
                 label="portfolio_value")
        plt.show()

    def test_get_open_dates(self):
        self.assertEqual(244, len(self.env.dates))

    def test_get_init_portfolio_obss(self):
        self.env.reset()
        actual = self.env.get_init_portfolio_obs()
        self.assertEqual(10, len(actual))
        self.assertEqual([0, 0, 0, 0], actual[0].tolist())

    def test_get_init_obs(self):
        self.env.reset()
        actual = self.env.get_init_obs()
        self.assertEqual(10, len(actual))
        self.assertEqual(24, len(actual[0]))
        # TODO: 更具体的测试

    def test_get_action_price(self):
        self.env.reset()
        # pre_close: 10.24
        action, code = [-0.1, 0.1], self.codes[0]
        sell_price, buy_price = self.env.get_action_price(action, code)
        self.assertEqual(10.14, sell_price)
        self.assertEqual(10.34, buy_price)

    def test_buy_and_hold(self):
        self.env.reset()
        action = self.env.get_buy_close_action(datestr=self.env.current_date)
        obs, reward, done, info, _ = self.env.step(action, only_update=False)
        action = [0] * 4
        while not done:
            # buy and hold, 持仓不动
            _, _, done, _, _ = self.env.step(action, only_update=True)
        self.assertEqual(144057.17, self.env.portfolio_value)
        if self.show_plot:
            self.plot_portfolio_value("buy_and_hold")

    def test_random(self):
        random.seed(0)
        self.env.reset()
        action = self.env.get_buy_close_action(datestr=self.env.current_date)
        obs, reward, done, info, _ = self.env.step(action, only_update=False)
        while not done:
            # buy and hold, 持仓不动
            action = self.env.get_random_action()
            _, _, done, _, _ = self.env.step(action, only_update=False)
        self.assertEqual(21408.2, round(self.env.portfolio_value, 1))
        if self.show_plot:
            self.plot_portfolio_value("random_action")

    def test_static(self):
        self.env.reset()
        action = self.env.get_buy_close_action(datestr=self.env.current_date)
        obs, reward, done, info, _ = self.env.step(action, only_update=False)
        action = [0.1, -0.1, 0.1, -0.1]
        while not done:
            # buy and hold, 持仓不动
            _, _, done, _, _ = self.env.step(action, only_update=False)
        self.assertEqual(104933, int(self.env.portfolio_value))
        if self.show_plot:
            self.plot_portfolio_value("static")

    def test_get_buy_close_action(self):
        actual = self.env.get_buy_close_action(datestr='20190116')
        expect = [0, 0.23438, 0, 0.10334000000000002]
        self.assertEqual(expect, actual)


if __name__ == '__main__':
    unittest.main()



import unittest

from tenvs.envs.reward import get_reward_func, mean_squared_error


class TestReward(unittest.TestCase):
    def test_mean_squared_error(self):
        a = [2.0, 4.0]
        b = [1.0, 4.0]
        mse = mean_squared_error(a, b)
        self.assertEqual(12.5, mse)

    def test_daily_return_add_price_bound(self):
        fn = get_reward_func("daily_return_add_price_bound")
        highs, lows, closes = [100], [90], [95]
        sell_prices, buy_prices = [98], [92]
        v = fn(0.03, highs, lows, closes, sell_prices, buy_prices)
        self.assertEqual(-0.05938271604938244, v)

    def test_daily_return_add_buy_sell_penalty(self):
        fn = get_reward_func("daily_return_add_buy_sell_penalty")
        highs, lows, closes = [100], [90], [95]
        sell_prices, buy_prices = [98], [92]
        v = fn(0.03, highs, lows, closes, sell_prices, buy_prices)
        self.assertEqual(0.09, v)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-

import numpy as np

from tenvs.envs.base import BaseEnv
from tenvs.logger import logger


class SimpleEnv(BaseEnv):
    """
    单支股票全量日内买卖
    action: [scaled_sell_price, scaled_buy_price], 取值[-1, 1], 对应[-0.1, 0.1]
    即: 出价较前一交易日的涨跌幅
    先以sell_price 卖出, 再以 buy_price 买进
    NOTE(wen): 实际交易时，可能与模拟环境存在差异
        1. 先到最高价，然后再到最低价：与模拟环境一致
        2. 先到最低价，再到最高价，这里出价有两种情况
            有现金，则按最低价买进，与模拟环境一致
            无现金(满仓)，模拟环境可以成交，实盘交易时的成交状态不一定可以成交
    """

    def __init__(self, market=None, investment=100000.0, look_back_days=10,
                 used_infos=["equities_hfq_info", "indexs_info"],
                 reward_fn="daily_return_add_price_bound", log_deals=False):
        """
        investment: 初始资金
        look_back_days: 向前取数据的天数
        """
        super(SimpleEnv, self).__init__(market, investment, look_back_days,
                                        used_infos, reward_fn, log_deals)
        # 股票数量
        self.n = 1
        self.action_space = 2
        self.code = market.codes[0]
        self.portfolio_info_size = 2
        self.input_size = self.market_info_size + self.portfolio_info_size

    def get_init_portfolio_obs(self):
        # 初始持仓信息
        self.portfolio = self.portfolios[0]
        one_day = np.array([self.portfolio.daily_return,
                            self.portfolio.value_percent])
        obs = np.array([one_day] * self.look_back_days)
        return obs

    def get_init_obs(self):
        """
        obs 由两部分组成: 市场信息, 帐户信息(收益率, 持仓量)
        """
        market_info = []
        for date in self.dates[: self.look_back_days]:
            market_info.append(self.get_market_info(date))
        market_info = np.array(market_info)
        portfolio_info = self.get_init_portfolio_obs()
        return np.concatenate((market_info, portfolio_info), axis=1)

    def get_action_price(self, action):
        pre_close = self.market.get_pre_close_price(
            self.code, self.current_date)
        logger.debug("%s %s pre_close: %.2f" %
                     (self.current_date, self.code, pre_close))
        [v_sell, v_buy] = action
        # scale [-1, 1] to [-0.1, 0.1]
        pct_sell, pct_buy = v_sell * 0.1, v_buy * 0.1
        sell_price = round(pre_close * (1 + pct_sell), 2)
        buy_price = round(pre_close * (1 + pct_buy), 2)
        return sell_price, buy_price

    def do_action(self, action, pre_portfolio_value, only_update):
        sell_price, buy_price = self.get_action_price(action)
        sell_prices, buy_prices = [sell_price], [buy_price]
        if only_update:
            sell_prices, buy_prices = [0] * self.n, [0] * self.n
        divide_rate = self.market.get_divide_rate(self.code, self.current_date)
        logger.debug("divide_rate: %.4f" % divide_rate)
        self.portfolio.update_before_trade(divide_rate)
        cash_change = 0

        if not only_update:
            # 全仓卖出
            sell_cash_change, ok = self.sell(0, sell_price, 0.0)
            # 全仓买进
            buy_cash_change, ok = self.buy(0, buy_price, 1.0)
            cash_change = buy_cash_change + sell_cash_change
            logger.debug("do_action: time_id: %d, %s, cash_change: %.1f" % (
                self.current_time_id, self.code, cash_change))

        close_price = self.market.get_close_price(self.code, self.current_date)
        self.portfolio.update_after_trade(
            close_price=close_price,
            cash_change=cash_change,
            pre_portfolio_value=pre_portfolio_value)
        return sell_prices, buy_prices

    def _next(self):
        market_info = self.get_market_info(self.current_date)
        portfolio_info = np.array([self.portfolio.daily_return,
                                  self.portfolio.value_percent])
        new_obs = np.concatenate((market_info, portfolio_info), axis=0)
        obs = np.concatenate((self.obs[1:, :],
                              np.array([new_obs])), axis=0)
        if not self.done:
            self.current_time_id += 1
            self.current_date = self.dates[self.current_time_id]
        self.pre_cash = self.cash
        return obs

    def get_buy_close_action(self, datestr):
        """
        在datestr以收盘价买进的action, 用于buy_and_hold策略, 给出的action为: [0, v]
        v为收盘价对应的action值
        """
        pct = self.market.codes_history[self.code].loc[datestr,  "pct_chg"]
        buy_act_v = self.scale_pct_to_action_value(pct)
        # NOTE: 卖出价格默认为：昨日收盘价
        sell_act_v = 0
        return [sell_act_v, buy_act_v]



# -*- coding:utf-8 -*-
import random
import time

import gym
import numpy as np

from tenvs.envs.reward import get_reward_func
from tenvs.logger import logger
from tenvs.portfolio import Portfolio


class BaseEnv(gym.Env):
    def __init__(self, market=None, investment=100000.0, look_back_days=10,
                 used_infos=["equities_hfq_info", "indexs_info"],
                 reward_fn="daily_return_add_price_bound", log_deals=False):
        """
        investment: 初始资金
        look_back_days: 向前取数据的天数
        """
        self.market = market
        # 股票数量
        self.n = len(market.codes)
        self.codes = market.codes
        self.start = market.start
        self.end = market.end
        self.look_back_days = look_back_days
        self.investment = investment
        # 输入数据: 个股信息 + 指数信息
        self.used_infos = used_infos
        self.market_info_size = self.get_market_info_size()
        # 开市日期列表
        self.dates = market.open_dates
        # 记录一个回合的收益序列
        self.returns = []
        self.reward_fn = get_reward_func(name=reward_fn)
        self.reward_fn_name = reward_fn
        self.log_deals = log_deals

    def get_market_info_size(self):
        size = 0
        for info_name in self.used_infos:
            size += self.market.get_info_size(info_name)
        return size

    def _init_current_time_id(self):
        return self.look_back_days

    def get_market_info(self, date):
        info = []
        for info_name in self.used_infos:
            data = self.market.market_info[date][info_name]
            info.extend(data)
        return np.array(info)

    def get_hlc_prices(self):
        date = self.current_date
        highs, lows, closes = [], [], []
        for i in range(self.n):
            code = self.codes[i]
            if date in self.market.codes_history[code].index:
                highs.append(self.market.codes_history[code].loc[date, "high"])
                lows.append(self.market.codes_history[code].loc[date, "low"])
                closes.append(self.market.codes_history[code].loc[date,
                                                                  "close"])
            else:
                # 停牌时
                highs.append(0)
                lows.append(0)
                closes.append(0)
        return highs, lows, closes

    def update_reward(self, sell_prices, buy_prices):
        if self.reward_fn_name in ["daily_return", "simple"]:
            self.reward = self.reward_fn(self.daily_return)
        else:
            highs, lows, closes = self.get_hlc_prices()
            self.reward = self.reward_fn(
                self.daily_return, highs, lows, closes,
                sell_prices, buy_prices)
        # 每一只股的reward与总的reward一致
        for i in range(self.n):
            self.rewards[i] = self.reward

    def sell(self, id, price, target_pct):
        # id: code id
        code = self.codes[id]
        logger.debug("sell %s, bid price: %.2f" % (code, price))
        ok, price = self.market.sell_check(
            code=code,
            datestr=self.current_date,
            bid_price=price)
        if ok:
            # 全仓卖出
            cash_change, price, vol = self.portfolios[
                id].order_target_percent(
                    percent=target_pct, price=price,
                    pre_portfolio_value=self.portfolio_value,
                    current_cash=self.cash)
            self.cash += cash_change
            if vol != 0:
                self.info["orders"].append(["sell", code,
                                            round(cash_change, 1),
                                            round(price, 2), vol])
            logger.debug("sell %s target_percent: 0, cash_change: %.3f" %
                         (code, cash_change))
            return cash_change, ok
        return 0, ok

    def buy(self, id, price, target_pct):
        # id: code id
        code = self.codes[id]
        logger.debug("buy %s, bid_price: %.2f" % (code, price))
        ok, price = self.market.buy_check(
            code=code,
            datestr=self.current_date,
            bid_price=price)
        pre_cash = self.cash
        if ok:
            # 分仓买进
            cash_change, price, vol = self.portfolios[id].order_target_percent(
                percent=target_pct, price=price,
                pre_portfolio_value=self.portfolio_value,
                current_cash=self.cash)
            self.cash += cash_change
            if vol != 0:
                self.info["orders"].append(["buy", code,
                                            round(cash_change, 1),
                                            round(price, 2), vol])
            logger.debug("buy %s cash: %.1f, cash_change: %1.f" %
                         (code, pre_cash, cash_change))
            return cash_change, ok
        return 0, ok

    def update_portfolio(self):
        pre_portfolio_value = self.portfolio_value
        self.market_value = 0
        self.daily_pnl = 0
        self.pnl = 0
        self.transaction_cost = 0
        self.all_transaction_cost = 0
        for p in self.portfolios:
            self.market_value += p.market_value
            self.daily_pnl += p.daily_pnl
            self.pnl += p.pnl
            self.transaction_cost += p.transaction_cost
            self.all_transaction_cost += p.all_transaction_cost
        self.total_pnl += self.pnl

        # 当日收益率 更新
        if pre_portfolio_value == 0:
            self.daily_return = 0
        else:
            self.daily_return = self.daily_pnl / pre_portfolio_value
        # update portfolio_value
        self.portfolio_value = self.market_value + self.cash
        self.portfolio_value_logs.append(self.portfolio_value)

    def update_value_percent(self):
        if self.portfolio_value == 0:
            self.value_percent = 0.0
        else:
            self.value_percent = self.market_value / self.portfolio_value
        for i in range(self.n):
            self.portfolios[i].update_value_percent(self.portfolio_value)

    def get_init_obs():
        raise NotImplementedError

    def reset(self):
        # 当前时间
        self.current_time_id = self._init_current_time_id()
        self.current_date = self.dates[self.current_time_id]
        self.done = False
        # 当日的回报
        self.reward = 0
        self.rewards = [0] * self.n
        # 累计回报
        self.total_reward = 0.0
        # 当日订单集合
        self.info = {"orders": []}
        # 总权益
        self.portfolio_value = self.investment
        # 初始资金
        self.starting_cash = self.investment
        # 可用资金
        self.cash = self.investment
        self.pre_cash = self.cash
        self.total_pnl = 0

        # 每只股的 portfolio
        self.portfolios = []
        for code in self.codes:
            self.portfolios.append(Portfolio(code=code,
                                             log_deals=self.log_deals))
        self.obs = self.get_init_obs()
        self.portfolio_value_logs = []
        return self.obs

    def step(self, action, only_update=False):
        """
        only_update为True时，表示buy_and_hold策略，可作为一种baseline策略
        """
        self.action = action
        self.info = {"orders": []}
        if self.log_deals:
            logger.info("=" * 50 + "%s" % self.current_date + "=" * 50)
        logger.debug("=" * 50 + "%s" % self.current_date + "=" * 50)
        logger.debug("current_time_id: %d, portfolio: %.1f" %
                     (self.current_time_id, self.portfolio_value))
        logger.debug("step action: %s" % str(action))

        # 到最后一天
        if self.current_date == self.dates[-1]:
            self.done = True

        pre_portfolio_value = self.portfolio_value
        sell_prices, buy_prices = self.do_action(action,
                                                 pre_portfolio_value,
                                                 only_update)
        self.update_portfolio()
        self.update_value_percent()
        self.update_reward(sell_prices, buy_prices)
        self.obs = self._next()
        self.info = {
            "orders": self.info["orders"],
            "current_date": self.current_date,
            "portfolio_value": round(self.portfolio_value / self.investment, 3),
            "daily_pnl": round(self.daily_pnl, 1),
            "reward": self.reward}
        return self.obs, self.reward, self.done, self.info, self.rewards

    def get_random_action(self):
        return [random.uniform(-1, 1) for i in range(self.action_space)]

    def scale_pct_to_action_value(self, pct):
        """
        将pct值scale到action值上, pct取值: [-10.0, 10.0], action的取值: [-1, 1]
        """
        return pct * 0.1



# -*- coding:utf-8 -*-

import logging
import os
import unittest

from tenvs.market import Market

logging.root.setLevel(logging.ERROR)


class TestMarket(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        # NOTE: 需要在环境变量中设置 TUSHARE_TOKEN
        ts_token = os.getenvs("TUSHARE_TOKEN")
        self.start = "20190101"
        self.end = "20200101"
        self.codes = ["000001.SZ"]
        self.indexs = ["000001.SH", "399001.SZ"]
        # self.indexs = []
        self.data_dir = "/tmp/tenvs"
        self.m = Market(
            ts_token=ts_token,
            start=self.start,
            end=self.end,
            codes=self.codes,
            indexs=self.indexs,
            data_dir=self.data_dir)

    def test_init_codes_history(self):
        self.assertEqual(244, len(self.m.indexs_history["000001.SH"]))
        self.assertEqual(244, len(self.m.codes_history[self.codes[0]]))

    def test_init_size_info(self):
        self.assertEqual(10, self.m.equity_hfq_info_size)
        self.assertEqual(18, self.m.indexs_info_size)

    def test_is_suspended(self):
        self.assertTrue(self.m.is_suspended(code='000', datestr=''))
        self.assertTrue(self.m.is_suspended(code='000001.SZ', datestr=''))
        # 星期六
        self.assertTrue(self.m.is_suspended(code='000001.SZ',
                                            datestr='20191012'))
        self.assertFalse(self.m.is_suspended(code='000001.SZ',
                                             datestr='20191021'))

    def test_buy_check(self):
        code = "000001.SZ"
        # is_suspended
        ok, price = self.m.buy_check(code='', datestr='')
        self.assertFalse(ok)
        self.assertEqual(0, price)
        datestr = "20191021"
        # print(self.m.codes_history["000001.SZ"].loc[datestr])
        # [high, low] = [16.97, 16.43]
        # 买入竞价低于最低价，不能成交
        ok, price = self.m.buy_check(code=code, datestr=datestr,
                                     bid_price=11)
        self.assertFalse(ok)

        # 买入竞价在最低最高价之间， 可以成交
        ok, price = self.m.buy_check(code=code, datestr=datestr,
                                     bid_price=16.51)
        self.assertTrue(ok)
        self.assertEqual(16.51, price)

        # 买入竞价高于最高介， 可以成交
        ok, price = self.m.buy_check(code=code, datestr=datestr,
                                     bid_price=17)
        self.assertTrue(ok)
        self.assertEqual(16.97, price)

    def test_sell_check(self):
        code = "000001.SZ"
        # is_suspended
        ok, price = self.m.sell_check(code='', datestr='')
        self.assertFalse(ok)
        self.assertEqual(0, price)
        datestr = "20191021"
        # [high, low] = [16.97, 16.43]
        # 卖出入竞价高于最高价，不能成交
        ok, price = self.m.sell_check(code=code, datestr=datestr,
                                      bid_price=17)
        self.assertFalse(ok)

        # 买入竞价在最低最高价之间， 可以成交
        ok, price = self.m.sell_check(code=code, datestr=datestr,
                                      bid_price=16.51)
        self.assertTrue(ok)
        self.assertEqual(16.51, price)

        # 卖出竞价低于最低价， 可以成交, 按最低价成交
        ok, price = self.m.sell_check(code=code, datestr=datestr, bid_price=16)
        self.assertTrue(ok)
        self.assertEqual(16.43, price)


if __name__ == '__main__':
    unittest.main()



# -*- coding:utf-8 -*-
import os

import pandas as pd
import tushare as ts

from tenvs.logger import logger


class Market:
    """
    模拟市场，加载环境所需要的数据
    NOTE(wen):
        从多数实验可以发现，使用后复权的数据做为模型输入，性能更优， Market统一使用
        后复权数据做模型的输入, 而不复权的数据做买卖的判断
    indexs:
        000001.SH: 上证指数
        399001.SZ: 深证成指
        ...
    NOTE(wen): 使用tushare下载指数日线信息需要在tushare.pro帐户中有200积分
    data_dir: 存储数据文件的目录，以降低重复下载的频率
    """

    def __init__(self,
                 ts_token="",
                 start="20190101",
                 end="20200101",
                 codes=["000001.SZ"],
                 indexs=["000001.SH", "399001.SZ"],
                 data_dir="/tmp/tenvs"):
        ts.set_token(ts_token)
        self.start = start
        self.end = end
        self.codes = codes
        self.indexs = indexs
        self.data_dir = data_dir
        self.load_codes_history()
        self.load_indexs_history()
        # hfq数据: 去除不复权数据(9列) 和复权因子(1列): 10, 从第11列开始是后复权数据
        self.equity_hfq_info_start_index = 10
        self.init_market_info()
        self.init_size_info()

    def get_info_size(self, info_name):
        date = self.open_dates[0]
        return len(self.market_info[date][info_name])

    def init_size_info(self):
        """
        初始化数据的size
        """
        self.equity_hfq_info_size = self.get_info_size("equities_hfq_info")
        self.indexs_info_size = self.get_info_size("indexs_info")

    def get_code_history(self, code, adj=None):
        return ts.pro_bar(
            ts_code=code, adj=adj,
            start_date=self.start, end_date=self.end)

    def load_codes_history(self):
        """
        self.codes_history: dict
        每条记录由两部分数据组成: 复权因子(1列), 不复权数据(9列), 后复权数据(9列)
            复权因子: adj_factor(1列)
            不复权OHLCV, ... (9列)
            后复权OHLCV, ... (9列)

        self.codes_history[code].columns的值为:
        [u'adj_factor', u'open', u'high', u'low', u'close', u'pre_close',
        u'change', u'pct_chg', u'vol', u'amount', u'open_hfq', u'high_hfq',
        u'low_hfq', u'close_hfq', u'pre_close_hfq', u'change_hfq',
        u'pct_chg_hfq', u'vol_hfq', u'amount_hfq']
        """

        self.codes_history = {}
        for code in self.codes:
            dir = os.path.join(self.data_dir, code)
            if not os.path.exists(dir):
                os.makedirs(dir)
            data_path = os.path.join(dir,
                                     self.start + "-" + self.end + ".csv")
            if os.path.exists(data_path):
                df = pd.read_csv(data_path)
                df = df.set_index("trade_date")
                df.index = df.index.astype(str, copy=False)
                self.codes_history[code] = df

            else:
                # 不复权
                df_bfq = self.get_code_history(code, adj=None)
                df_bfq = df_bfq.drop(columns=["ts_code"], axis=1)
                # 后复权
                df_hfq = self.get_code_history(code, adj="hfq")
                df_hfq = df_hfq.drop(columns=["ts_code"], axis=1)
                df = pd.merge(df_bfq, df_hfq,
                              on='trade_date', how='left',
                              suffixes=('', '_hfq'))
                # 拆分因子
                col_name = df.columns.tolist()
                col_name.insert(0, 'adj_factor')
                df = df.reindex(columns=col_name)
                df["adj_factor"] = df["close_hfq"] / df["close"]
                df = df.sort_values(by="trade_date", ascending=True)
                df = df.set_index("trade_date")
                df.index = df.index.astype(str, copy=False)
                df.to_csv(data_path)
                self.codes_history[code] = df

    def load_indexs_history(self):
        self.indexs_history = {}
        # 默认加载: 000001.SH(上证指数), 399001.SZ(深城证指)
        indexs = ["000001.SH", "399001.SZ"]
        for code in self.indexs:
            if code not in indexs:
                indexs.append(code)

        for code in indexs:
            dir = os.path.join(self.data_dir, "indexs", code)
            if not os.path.exists(dir):
                os.makedirs(dir)
            data_path = os.path.join(dir,
                                     self.start + "-" + self.end + ".csv")
            if os.path.exists(data_path):
                df = pd.read_csv(data_path)
                df = df.set_index("trade_date")
                df.index = df.index.astype(str, copy=False)
                self.indexs_history[code] = df
            else:
                pro = ts.pro_api()
                df = pro.index_daily(ts_code=code,
                                     start_date=self.start,
                                     end_date=self.end)
                df = df.drop(columns=["ts_code"], axis=1)
                df = df.sort_values(by="trade_date", ascending=True)
                df = df.set_index("trade_date")
                df.index = df.index.astype(str, copy=False)
                df.to_csv(data_path)
                self.indexs_history[code] = df

    def set_pre_info(self):
        date = self.open_dates[0]
        self.equities_pre_hfq_info = {}
        self.equities_pre_bfq_info = {}
        for code in self.codes:
            # 如果第一天就停牌，这里会出错，建议另外选择一天开始回测
            try:
                data = self.codes_history[code].loc[date].tolist()
                bfq_data = data[1: self.equity_hfq_info_start_index]
                # 开盘时， open=1
                bfq_data.append(1)
                self.equities_pre_bfq_info[code] = bfq_data
                # 后复权
                hfq_data = data[self.equity_hfq_info_start_index:]
                # 开盘时， open=1
                hfq_data.append(1)
                self.equities_pre_hfq_info[code] = hfq_data
            except Exception as e:
                print("%s, %s停牌，建议另外选择一天开始回测" % (code, date))
                print(e)
                exit()

    def get_equities_bfq_info(self, date):
        info = []
        for code in self.codes:
            data = None
            if date in self.codes_history[code].index:
                data = self.codes_history[code].loc[date].tolist()
                data = data[1: self.equity_hfq_info_start_index]
                # 开盘时， open=1
                data.append(1)
            else:
                # 如果停牌则使用前一开盘日信息
                data = self.equities_pre_bfq_info[code]
                data[-1] = 0
            # 更新前一开盘日信息
            self.equities_pre_bfq_info[code] = data
            info.extend(data)
        return info

    def get_equities_hfq_info(self, date):
        info = []
        for code in self.codes:
            data = None
            if date in self.codes_history[code].index:
                data = self.codes_history[code].loc[date].tolist()
                data = data[self.equity_hfq_info_start_index:]
                # 开盘时， open=1
                data.append(1)
            else:
                # 如果停牌则使用前一开盘日信息
                data = self.equities_pre_hfq_info[code]
                data[-1] = 0
            # 更新前一开盘日信息
            self.equities_pre_hfq_info[code] = data
            info.extend(data)
        return info

    def get_indexs_info(self, date):
        info = []
        for code in self.indexs:
            data = self.indexs_history[code].loc[date].tolist()
            info.extend(data)
        return info

    def init_market_info(self):
        """
        以日期为key, 将市场相关信息组织在一个dict中:
            equities_hfq_info: 合并之后的个股信息, 如果某支股停牌，则使用它前一开盘日信息
            indexs_info: 合并之后指数信息
        Note(wen): 添加其他市场相关信息，都放在这里
        """
        self.open_dates = self.indexs_history["000001.SH"].index.tolist()
        self.open_dates.sort()
        self.set_pre_info()
        self.market_info = {}
        for date in self.open_dates:
            self.market_info[date] = {}
            self.market_info[date]["equities_bfq_info"] = \
                self.get_equities_bfq_info(date)
            self.market_info[date]["equities_hfq_info"] = \
                self.get_equities_hfq_info(date)
            self.market_info[date]["indexs_info"] = self.get_indexs_info(date)

    def is_suspended(self, code='', datestr=''):
        # 是否停牌，是：返回 True, 否：返回 False
        if code not in self.codes_history:
            return True
        if datestr not in self.codes_history[code].index:
            return True
        return False

    def buy_check(self, code='', datestr='', bid_price=None):
        # 返回：OK, 成交价
        ok = False
        # 停牌
        if self.is_suspended(code, datestr):
            return ok, 0
        # 获取当天标的信息
        [open, high, low, pct_change] = self.codes_history[code].loc[
            datestr, ["open", "high", "low", "pct_chg"]]
        # 涨停封板, 无法买入
        if low == high and pct_change > self.top_pct_change:
            logger.debug(u"sell_check %s %s 涨停法买进" % (code, datestr))
            return ok, 0
        # 买入竞价低于最低价，不能成交
        if bid_price < low:
            return ok, 0
        # 买入竞价高于最低价， 可以成交
        if bid_price >= low:
            return True, min(bid_price, high)

    def sell_check(self, code='', datestr='', bid_price=None):
        # 返回：OK, 成交价
        ok = False
        # 停牌
        if self.is_suspended(code, datestr):
            return ok, 0
        # 获取当天标的信息
        [open, high, low, pct_change] = self.codes_history[code].loc[
            datestr, ["open", "high", "low", "pct_chg"]]
        # 跌停封板， 不能卖出
        if low == high and pct_change < -self.top_pct_change:
            logger.debug(u"sell_check %s %s 跌停无法卖出" % (code, datestr))
            return ok, 0
        # 卖出竞价高于最高价，不可以成交
        if bid_price > high:
            return ok, 0
        # 卖出竞价在最低最高价之间， 可以成交，按出价成交
        # NOTE: 这里卖出竞价低于最低价时，可以成交，按最低价成交
        if bid_price <= high:
            ok = True
            return ok, max(bid_price, low)

    def get_pre_close_price(self, code, datestr):
        if datestr in self.codes_history[code].index:
            return self.codes_history[code].loc[datestr]["pre_close"]
        # 如果当天停牌, 返回前一开市日的pre_close
        df = self.codes_history[code]
        df_ = df[df.index < datestr]
        return df_.iloc[-1]["pre_close"]

    def get_close_price(self, code, datestr):
        if datestr in self.codes_history[code].index:
            return self.codes_history[code].loc[datestr]["close"]
        # 如果当天停牌, 返回前一开市日收盘价
        df = self.codes_history[code]
        df_ = df[df.index < datestr]
        return df_.iloc[-1]["close"]

    def get_pre_adj_factor(self, code, datestr):
        df = self.codes_history[code]
        df_ = df[df.index < datestr]
        return df_.iloc[-1]["adj_factor"]

    def get_adj_factor(self, code, datestr):
        if self.is_suspended(code=code, datestr=datestr):
            return self.get_pre_adj_factor(code, datestr)
        else:
            return self.codes_history[code].loc[datestr]["adj_factor"]

    def get_divide_rate(self, code, datestr):
        pre_adj_factor = self.get_pre_adj_factor(code, datestr)
        current_adj_factor = self.get_adj_factor(code, datestr)
        return current_adj_factor / pre_adj_factor



# -*- coding:utf-8 -*-

from tenvs.logger import logger


class Portfolio:
    """
    NOTE: 考虑到不同的类型的资产commission并不一样, 所以Portfolio需要commission相关的参
    数
    divide_rate_threshold: 是否有拆分判断阀值
    """

    def __init__(self,
                 buy_commission_rate=0.001, sell_commission_rate=0.0015,
                 min_commission=5.0, round_lot=100,
                 divide_rate_threshold=1.005, code="000001.SZ",
                 log_deals=False):
        # 市值
        self.market_value = 0.0
        # 持仓量
        self.volume = 0
        # 前一持仓量
        self.pre_volume = 0
        # 冻结量
        self.frozen_volume = 0
        # 该仓位可卖出股数。T＋1 的市场中sellable = 所有持仓 - 已冻结
        self.sellable = 0
        # 平均开仓价格
        self.avg_price = 0.0
        # 最近一次成交的价格
        self._price = 0.0
        # 当日盈亏
        self.daily_pnl = 0.0
        # 累计盈亏, Profit and Loss
        self.pnl = 0.0
        # 当前最新一天的日收益率, 以总账户为基线
        self.daily_return = 0.0
        # 当日交易费
        self.transaction_cost = 0.0
        # 累计交易费
        self.all_transaction_cost = 0.0
        # 获得该持仓的市场价值在股票投资组合价值中所占比例，取值范围[0, 1], 以总账户为基线
        self.value_percent = 0.0
        # 是否停牌, 停牌为1
        self.is_suspended = 0.0
        # 买入佣金率
        self.buy_commission_rate = buy_commission_rate
        # 卖佣金率
        self.sell_commission_rate = sell_commission_rate
        # 单笔最低佣金
        self.min_commission = min_commission
        # 一手对应多少股，中国A股一手是100股
        self.round_lot = round_lot
        self.divide_rate_threshold = divide_rate_threshold
        self.code = code
        # 是否打印订单记录
        self.log_deals = log_deals

    def update_value_percent(self, total_value):
        """
        [float] 获得该持仓的实时市场价值在股票投资组合价值中所占比例，取值范围[0, 1]
        """
        if total_value == 0:
            self.value_percent = 0
        else:
            self.value_percent = self.market_value / total_value

    def _buy_fee(self, amount):
        # 1.印花税：由卖方出
        # 2.证管费：约为成交金额的0.00002收取
        # 3.证券交易经手费：A股，按成交金额的0.00696%收取
        # A股2、3项收费合计称为交易规费，合计收取成交金额的0.00896%，包含在券商交易佣金中。
        # 4.过户费: 按成交金额的0.002%收取。
        # 5.券商交易佣金：最高不超过成交金额的3‰，最低5元起，单笔交易佣金不满5元按5元收取。
        # NOTE: 实际操作费率约： 0.0010775285
        return round(max(self.min_commission,
                         amount * self.buy_commission_rate), 2)

    def _sell_fee(self, amount):
        # 1.印花税：成交金额的 0.001, 2008年9月19日至今由向双边征收改为向出让方单边征收
        # 2.证管费：约为成交金额的0.00002收取
        # 3.证券交易经手费：A股，按成交金额的0.00696%收取
        # A股2、3项收费合计称为交易规费，合计收取成交金额的0.00896%，包含在券商交易佣金中。
        # 4.过户费: 按成交金额的0.002%收取。
        # 实际操作费率： 0.00126019
        return round(amount * self.sell_commission_rate, 2)

    def buy(self, price, volume):
        # 买入
        amount = volume * price
        transaction_cost = self._buy_fee(amount)
        self.transaction_cost += transaction_cost
        # 平均开仓价更新
        self.avg_price = (self.avg_price * self.volume +
                          amount + transaction_cost) / (self.volume + volume)
        self._price = price
        self.volume += volume
        self.frozen_volume += volume
        self.all_transaction_cost += transaction_cost
        cash_change = -amount - transaction_cost
        if self.log_deals and volume > 0:
            logger.info("buy %s price: %.3f, volume: %d" % (self.code,
                                                            price, volume))
        logger.debug("buy %s price: %.3f, volume: %d" % (self.code,
                                                         price, volume))
        return cash_change, price, volume

    def is_divide(self, divide_rate):
        # 拆分判断, 有拆分返回 True, 否则 False
        return divide_rate > self.divide_rate_threshold

    def update_before_trade(self, divide_rate):
        # 如果有拆分
        if self.is_divide(divide_rate):
            logger.debug("update_before_trade: %s divide_rate: %.3f" % (
                self.code, divide_rate))
            self.volume = int(divide_rate * self.volume)
        self.sellable = self.volume
        self.frozen_volume = 0
        self.daily_pnl = 0.0
        self.daily_return = 0.0
        self.transaction_cost = 0.0
        self.pre_volume = self.volume

    def sell(self, price, volume):
        # 卖出
        amount = volume * price
        transaction_cost = self._sell_fee(amount)
        # 平均开仓价更新
        if self.volume == volume:
            self.avg_price = 0.0
        else:
            self.avg_price = (self.avg_price * self.volume - amount +
                              transaction_cost) / (self.volume - volume)
        self._price = price
        self.transaction_cost += transaction_cost
        self.volume -= volume
        self.sellable -= volume
        self.all_transaction_cost += transaction_cost
        cash_change = amount - transaction_cost
        if self.log_deals and volume > 0:
            logger.info("sell %s price: %.3f, volume: %d" % (self.code,
                                                             price, volume))
        logger.debug("sell %s price: %.3f, volume: %d" % (self.code,
                                                          price, volume))
        return cash_change, price, volume

    def _submit_order(self, side=None, price=None, volume=None):
        logger.debug("portfolio: submit_order %s, %4s, %5.2f, %6d" % (
            self.code, side, price, volume))
        if side == "sell":
            return self.sell(price=price, volume=volume)
        elif side == "buy":
            return self.buy(price=price, volume=volume)

    def _sell_all_stock(self, price):
        if self.sellable == 0:
            logger.debug("%s sell_all_stock sellable is 0" % self.code)
        return self._submit_order(side="sell", price=price,
                                  volume=self.sellable)

    def order_value(self, amount, price, current_cash):
        """
        使用想要花费的金钱买入/卖出股票，而不是买入/卖出想要的股数，正数代表买入，负数代表卖出
        股票的股数总是会被调整成对应的100的倍数（在A中国A股市场1手是100股）
        如果资金不足，将不会创建发送订单
        需要注意：
        当您提交一个买单时，cash_amount 代表的含义是您希望买入股票消耗的金额（包含税费），
        最终买入的股数不仅和发单的价格有关，还和税费相关的参数设置有关。
        当您提交一个卖单时，cash_amount 代表的意义是您希望卖出股票的总价值。
        如果金额超出了您所持有股票的价值，那么您将卖出所有股票。
        :example:
        .. code-block:: python
            #花费最多￥10000买入平安银行股票，并以市价单发送。具体下单的数量与您策略税费相关
            的配置有关。
            order_value('000001.XSHE', 10000)
            #卖出价值￥10000的现在持有的平安银行：
            order_value('000001.XSHE', -10000)
        """
        if amount > 0:
            amount = min(amount, current_cash)
            volume = int(amount / (price * self.round_lot)) * self.round_lot
            while volume > 0:
                amount = volume * price
                expected_transaction_cost = self._buy_fee(amount)
                if amount + expected_transaction_cost <= current_cash:
                    break
                volume -= self.round_lot
            if volume > 0:
                return self._submit_order("buy", price, volume)
            if volume == 0:
                logger.debug("order_value failed: 0 order quantity")
                return 0, 0, 0

        elif amount < 0:
            vol = int(amount / (price * self.round_lot)) * self.round_lot
            volume = abs(vol)
            # 根据可卖出股数, 按最大可交易量，调整
            volume = min(self.sellable, volume)
            return self._submit_order(side="sell", price=price, volume=volume)
        else:
            return 0, 0, 0

    def order_target_percent(self, percent=None, price=None,
                             pre_portfolio_value=None, current_cash=None):
        """
        NOTE: 在调用之前，应该确定price是可以交易的，这里不对交易量进行限制.
        买入/卖出证券以自动调整该证券的仓位到占有一个目标价值.
        加仓时，percent 代表证券已有持仓的价值加上即将花费的现金（包含税费）的总值占当前投资
        组合总价值的比例。
        减仓时，percent 代表证券将被调整到的目标价至占当前投资组合总价值的比例。
        投资组合价值等于所有已有仓位的价值和剩余现金的总和。
        买/卖单会被下舍入一手股数（A股是100的倍数）的倍数。目标百分比应该是一个小数，
        并且最大值应该<=1，比如0.5表示50%。
        如果position_to_adjust 计算之后是正的，那么会买入该证券，否则会卖出该证券.
        需要注意，如果资金不足，该API将不会创建发送订单.
        """
        if percent < 0 or percent > 1:
            raise Exception(u"percent should between 0 and 1")
            return 0

        if percent == 0:
            return self._sell_all_stock(price)
        # portfolio_value 为上一交易日的值，并没有在每笔交易时更新，
        # 所以实时交易时，这里需要调整为实时的 portfolio_value
        adjust = pre_portfolio_value * percent - self.volume * self._price
        return self.order_value(adjust, price, current_cash)

    def update_after_trade(self, close_price, cash_change,
                           pre_portfolio_value):
        # 昨日收盘市值
        pre_market_value = self.market_value
        # 收盘后，根据收盘价更新
        self.market_value = self.volume * close_price
        self.daily_pnl = self.market_value - pre_market_value + cash_change
        self.pnl += self.daily_pnl

        if pre_portfolio_value == 0:
            self.daily_return = 0
        else:
            self.daily_return = self.daily_pnl / pre_portfolio_value



# -*- coding:utf-8 -*-
from tenvs.logger import logger

mapping = {}


def register(name):
    def _thunk(func):
        mapping[name] = func
        return func
    return _thunk


@register("simple")
def simple(daily_return, *args):
    if daily_return <= 0:
        return -1
    else:
        return 1


def daily_return(daily_return, *args):
    return daily_return


@register("daily_return_add_count_rate")
def daily_return_add_count_rate(daily_return, highs, lows,
                                closes, sell_prices, buy_prices):
    fail, success, profit_count, loss_count = 0, 0, 0
    for i in range(len(highs)):
        # 买
        if buy_prices[i] >= lows[i]:
            success += 1
            if buy_prices[i] <= closes[i]:
                profit_count += 1
            else:
                loss_count += 1
        else:
            fail += 1

        # 卖
        if sell_prices[i] <= highs[i]:
            success += 1
            if sell_prices[i] <= closes[i]:
                loss_count += 1
            else:
                profit_count += 1
        else:
            fail += 1

    success_rate = (success * 2) / (success + fail)
    profit_rate = (profit_count * 2) / (profit_count + loss_count)

    reward = daily_return + success_rate + profit_rate

    return reward


def mean_squared_error(a, b, scaled=10.0):
    """
    arguments:
        a: list of values
        b: list of values
        scaled: 误差调整倍率
    return:
        mean squared error
    """
    v = 0.0
    n = len(a)
    for i in range(n):
        if a[i] != 0:
            v += (scaled * (1 - b[i] / a[i])) ** 2
    return v / n


@register("daily_return_add_buy_sell_penalty")
def daily_return_add_buy_sell_penalty(daily_return, highs, lows,
                                      closes, sell_prices, buy_prices):
    reward = daily_return
    n = len(highs)
    # 如果出现买价>卖价 增加一个较大的惩罚
    for i in range(n):
        if sell_prices[i] < buy_prices[i]:
            reward -= 0.05
    reward = reward * 3
    return reward


@register("daily_return_add_price_bound")
def daily_return_add_price_bound(daily_return, highs, lows,
                                 closes, sell_prices, buy_prices):
    reward = daily_return
    n = len(highs)
    # 如果出现买价>卖价 增加一个较大的惩罚
    for i in range(n):
        if sell_prices[i] < buy_prices[i]:
            reward -= 1.0
    # 计算 bound
    sell_error = mean_squared_error(highs, sell_prices)
    buy_error = mean_squared_error(lows, buy_prices)

    reward = reward - sell_error - buy_error
    return reward


@register("daily_return_with_chl_penalty")
def daily_return_with_chl_penalty(daily_return, highs, lows,
                                  closes, sell_prices, buy_prices):
    reward = daily_return_add_price_bound(daily_return, highs, lows, closes,
                                          sell_prices, buy_prices)
    # 增加相对于收盘价的惩罚
    close_error_sum = 0
    n = len(highs)
    for i in range(n):
        if closes[i] == 0:
            continue
        if sell_prices[i] < closes[i]:
            close_error = (closes[i] - sell_prices[i]) * 10 / closes[i]
            close_error_sum += close_error ** 2
        if buy_prices[i] > closes[i]:
            close_error = (buy_prices[i] - closes[i]) * 10 / closes[i]
            close_error_sum += close_error ** 2
    reward = reward + close_error_sum
    return reward


def get_reward_func(name):
    """
    If you want to register your own reward function, you just need:
    Usage Example:
    -------------
    from tenvs.envs.reward import register
    @register("your_reward_function_name")
    def your_reward_function(**kwargs):
        ...
        return reward_fn
    """
    logger.info("tenvs.envs.reward use reward function: %s" % name)
    if callable(name):
        return name
    elif name in mapping:
        return mapping[name]
    else:
        raise ValueError('Unknown network type: {}'.format(name))


def main():
    r_func = get_reward_func(name="simple")
    assert -1 == r_func(0)
    assert 1 == r_func(0.01)
    assert -1 == r_func(-0.01)


if __name__ == '__main__':
    main()



from setuptools import find_packages, setup


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='tenvs',
      version='1.0.0',
      description='',
      url='https://github.com/iminders/tenvs',
      author='liuwen',
      author_email='liuwen.w@qq.com',
      long_description=long_description,
      long_description_content_type="text/markdown",
      packages=find_packages(),
      include_package_data=True,
      zip_safe=False,
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        ],
      python_requires='>=3',
      install_requires=[]
      )



# -*- coding:utf-8 -*-
from tenvs.envs.average import AverageEnv
from tenvs.envs.multi_vol import MultiVolEnv
from tenvs.envs.simple import SimpleEnv


def make_env(scenario, market, investment, look_back_days,
             used_infos, reward_fn, log_deals):
    if scenario == "simple":
        return SimpleEnv(market, investment, look_back_days,
                         used_infos, reward_fn, log_deals)
    elif scenario == "average":
        return AverageEnv(market, investment, look_back_days,
                          used_infos, reward_fn, log_deals)
    elif scenario == "multi_vol":
        return MultiVolEnv(market, investment, look_back_days,
                           used_infos, reward_fn, log_deals)
    else:
        raise "Not implement scenario %S" % scenario
